package cypher

import (
	"encoding/json"
	"fmt"
	"testing"

	"git.science.uu.nl/graphpolaris/query-conversion/entity"
)

func Test1(t *testing.T) {
	// Works, but, the AVG function is applied to a string, so that doesnt work, but the translation does :D
	query := []byte(`{
		"databaseName": "Movies3",
		"return": {
			"entities": [
				0,
				1,
				2
			],
			"relations": [
				0,
				1
			],
			"groupBys": [
				0
			]
		},
		"entities": [
			{
				"id": 0,
				"name": "Person",
				"constraints": [
				{
					"attribute": "name",
					"value": "Raymond Campbell",
					"dataType": "string",
					"matchType": "NEQ",
					"inID": -1,
					"inType": ""
				}
				]
			},
			{
				"id": 1,
				"name": "Movie",
				"constraints": []
			},
			{
				"id": 2,
				"name": "Genre",
				"constraints": []
			}
		],
		"relations": [
			{
				"id": 0,
				"name": "DIRECTED",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromID": 0,
				"toType": "entity",
				"toID": 1,
				"constraints": []
			},
			{
				"id": 1,
				"name": "IN_GENRE",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "groupBy",
				"fromID": 0,
				"toType": "entity",
				"toID": 2,
				"constraints": []
			}
		],
		"groupBys": [
			{
				"id": 0,
				"groupType": "entity",
				"groupID": 0,
				"groupAttribute": "bornIn",
				"byType": "entity",
				"byID": 1,
				"byAttribute": "imdbId",
				"appliedModifier": "AVG",
				"relationID": 0,
				"constraints": []
			}
		],
		"machineLearning": [],
		"limit": 5000
	}`)

	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	// hierarchy, err := createQueryHierarchy(&JSONQuery)
	// if err != nil {
	// 	fmt.Println(err)
	// }

	// fmt.Println(hierarchy)

	s := NewService()
	cypher, err := s.ConvertQuery2(&JSONQuery)
	if err != nil {
		fmt.Println(err)
	}

	fmt.Println(*cypher)

	t.Fail()

}
func Test2(t *testing.T) {
	// Works, but, the AVG function is applied to a string, so that doesnt work, but the translation does :D
	query := []byte(`{
		"databaseName": "TweedeKamer",
		"return": {
			"entities": [
				0,
				1,
				2
			],
			"relations": [
				0,
				1
			]
		},
		"entities": [
			{
				"name": "parliament",
				"ID": 0,
				"constraints": [
					{
						"attribute": "name",
						"value": "Geert",
						"dataType": "string",
						"matchType": "contains"
					}
				]
			},
			{
				"name": "parties",
				"ID": 1,
				"constraints": []
			},
			{
				"name": "resolutions",
				"ID": 2,
				"constraints": []
			}
		],
		"relations": [
			{
				"ID": 0,
				"name": "member_of",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromID": 0,
				"toType": "entity",
				"toID": 1,
				"constraints": []
			},
			{
				"ID": 1,
				"name": "submits",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromID": 0,
				"toType": "entity",
				"toID": 2,
				"constraints": []
			}
		],
		"groupBys": [],
		"machineLearning": [],
		"limit": 5000
	}
	`)

	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	s := NewService()
	cypher, err := s.ConvertQuery2(&JSONQuery)
	if err != nil {
		fmt.Println(err)
	}

	fmt.Println(*cypher)

	t.Fail()

}
func Test3(t *testing.T) {
	// Works, but, the AVG function is applied to a string, so that doesnt work, but the translation does :D
	query := []byte(`{
		"databaseName": "TweedeKamer",
		"return": {
			"entities": [
				0,
				1,
				2,
				3,
				4
			],
			"relations": [
				0,
				1,
				2,
				3
			]
		},
		"entities": [
			{
				"name": "parliament",
				"ID": 0,
				"constraints": [
					{
						"attribute": "name",
						"value": "A",
						"dataType": "string",
						"matchType": "contains"
					}
				]
			},
			{
				"name": "parties",
				"ID": 1,
				"constraints": [
					{
						"attribute": "seats",
						"value": "10",
						"dataType": "int",
						"matchType": "LT"
					}
				]
			},
			{
				"name": "resolutions",
				"ID": 2,
				"constraints": [
					{
						"attribute": "date",
						"value": "mei",
						"dataType": "string",
						"matchType": "contains"
					}
				]
			},
			{
				"name": "parliament",
				"ID": 3,
				"constraints": []
			},
			{
				"name": "parties",
				"ID": 4,
				"constraints": [
					{
						"attribute": "name",
						"value": "Volkspartij voor Vrijheid en Democratie",
						"dataType": "string",
						"matchType": "=="
					}
				]
			}
		],
		"relations": [
			{
				"ID": 0,
				"name": "member_of",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromID": 0,
				"toType": "entity",
				"toID": 1,
				"constraints": []
			},
			{
				"ID": 1,
				"name": "submits",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromID": 0,
				"toType": "entity",
				"toID": 2,
				"constraints": []
			},
			{
				"ID": 2,
				"name": "submits",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromID": 3,
				"toType": "entity",
				"toID": 2,
				"constraints": []
			},
			{
				"ID": 3,
				"name": "member_of",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromID": 3,
				"toType": "entity",
				"toID": 4,
				"constraints": []
			}
		],
		"groupBys": [],
		"machineLearning": [],
		"limit": 5000
	}
	
	`)

	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	s := NewService()
	cypher, err := s.ConvertQuery2(&JSONQuery)
	if err != nil {
		fmt.Println(err)
	}

	fmt.Println(*cypher)

	t.Fail()

}

func Test4(t *testing.T) {
	// Works, but, the AVG function is applied to a string, so that doesnt work, but the translation does :D
	query := []byte(`{
		"return": {
			"entities": [
				0,
				1,
				2,
				3,
				4,
				5,
				6,
				7
			],
			"relations": [
				0,
				1,
				2,
				3,
				4,
				5,
				6
			]
		},
		"entities": [
			{
				"name": "parliament",
				"ID": 0,
				"constraints": [
					{
						"attribute": "name",
						"value": "Geert",
						"dataType": "string",
						"matchType": "contains"
					}
				]
			},
			{
				"name": "commissions",
				"ID": 1,
				"constraints": []
			},
			{
				"name": "parliament",
				"ID": 2,
				"constraints": []
			},
			{
				"name": "parties",
				"ID": 3,
				"constraints": [
					{
						"attribute": "seats",
						"value": "10",
						"dataType": "int",
						"matchType": "LT"
					}
				]
			},
			{
				"name": "resolutions",
				"ID": 4,
				"constraints": [
					{
						"attribute": "date",
						"value": "mei",
						"dataType": "string",
						"matchType": "contains"
					}
				]
			},
			{
				"name": "resolutions",
				"ID": 5,
				"constraints": []
			},
			{
				"name": "parties",
				"ID": 6,
				"constraints": []
			}
			,
			{
				"name": "parliament",
				"ID": 7,
				"constraints": []
			}
		   
		],
		"groupBys": [],
		"relations": [
			{
				"ID": 0,
				"name": "part_of",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromId": 0,
				"toType": "entity",
				"toID": 1,
				"constraints": []
			},
			{
				"ID": 1,
				"name": "part_of",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromId": 2,
				"toType": "entity",
				"toID": 1,
				"constraints": []
			},
			{
				"ID": 2,
				"name": "member_of",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromId": 2,
				"toType": "entity",
				"toID": 3,
				"constraints": []
			},
			{
				"ID": 3,
				"name": "submits",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromId": 2,
				"toType": "entity",
				"toID": 4,
				"constraints": []
			},
			{
				"ID": 4,
				"name": "submits",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromId": 0,
				"toType": "entity",
				"toID": 5,
				"constraints": []
			},
			{
				"ID": 5,
				"name": "member_of",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromId": 0,
				"toType": "entity",
				"toID": 6,
				"constraints": []
			}
			,
			{
				"ID": 6,
				"name": "member_of",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromId": 7,
				"toType": "entity",
				"toID": 6,
				"constraints": []
			}
		],
		"machineLearning": [],
		"limit": 5000
	}
	
	
	`)

	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	s := NewService()
	cypher, err := s.ConvertQuery2(&JSONQuery)
	if err != nil {
		fmt.Println(err)
	}

	fmt.Println(*cypher)

	t.Fail()

}
func Test5(t *testing.T) {
	// Works, but, the AVG function is applied to a string, so that doesnt work, but the translation does :D
	query := []byte(`{
		"databaseName": "Movies3",
		"entities": [
			{
				"id": 0,
				"name": "Person",
				"constraints": []
			},
			{
				"id": 1,
				"name": "Movie",
				"constraints": []
			},
			{
				"id": 2,
				"name": "Person",
				"constraints": [
					{
						"attribute": "bornIn",
						"value": "",
						"dataType": "string",
						"matchType": "",
						"inID": 0,
						"inType": "groupBy"
					}
				]
			}
		],
		"relations": [
			{
				"id": 0,
				"name": "ACTED_IN",
				"depth": {
					"min": 1,
					"max": 1
				},
				"fromType": "entity",
				"fromID": 0,
				"toType": "entity",
				"toID": 1,
				"constraints": []
			}
		],
		"groupBys": [
			{
				"id": 0,
				"groupType": "entity",
				"groupID": 1,
				"groupAttribute": "imdbRating",
				"byType": "entity",
				"byID": 0,
				"byAttribute": "bornIn",
				"appliedModifier": "AVG",
				"relationID": 0,
				"constraints": [
					{
						"attribute": "imdbRating",
						"value": "7.5",
						"dataType": "int",
						"matchType": "GT",
						"inID": -1,
						"inType": ""
					}
				]
			}
		],
		"machineLearning": [],
		"limit": 5000
	}
	`)

	var JSONQuery entity.IncomingQueryJSON
	json.Unmarshal(query, &JSONQuery)

	s := NewService()
	cypher, err := s.ConvertQuery2(&JSONQuery)
	if err != nil {
		fmt.Println(err)
	}

	fmt.Println(*cypher)

	t.Fail()

}
